import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.distributions.normal import Normal
from torch.distributions.categorical import Categorical
from torch.distributions.mixture_same_family import MixtureSameFamily
import numpy as np
import copy,os
import scipy.stats
from collections import OrderedDict

def crt_net(hs,dropout=0.0,activation=0,bias=True,bn=0):
    net = []
    for h0,h1 in zip(hs, hs[1:]):
        if bn == 1:
            net.extend([nn.Linear(h0, h1,bias=bias),nn.BatchNorm1d(h1),nn.ReLU(),nn.Dropout(p=dropout)])
        else:
            net.extend([nn.Linear(h0, h1,bias=bias),nn.ReLU(),nn.Dropout(p=dropout)])            
    n = 2
    if bn == 1:
        n = 3
    for i in range(n):
        net.pop()
    return nn.Sequential(*net)

def crt_posnet(hs,dropout=0.0):
    net = []
    for h0,h1 in zip(hs, hs[1:]):
        net.extend([PositiveLinear(h0, h1),nn.Tanh(),nn.Dropout(p=dropout)])            
    n = 2
    for i in range(n):
        net.pop()
    return nn.Sequential(*net)

class PositiveLinear(nn.Module):
    def __init__(self, in_features, out_features):
        super(PositiveLinear, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.log_weight = nn.Parameter(torch.Tensor(out_features, in_features))
        self.bias = nn.Parameter(torch.zeros(out_features))
        self.reset_parameters()
    def reset_parameters(self):
        nn.init.xavier_uniform_(self.log_weight)
    def forward(self, input):
        return nn.functional.linear(input, self.log_weight.exp())+self.bias

class SparseGP(nn.Module):
    def __init__(self,n_att,n_out,n_idc=256,n_rank=32):
        super(SparseGP,self).__init__()
        self.y_idc = nn.Parameter(torch.Tensor(1,n_idc,n_out))
        self.x_idc = nn.Parameter(torch.Tensor(1,n_idc,n_att)) #inducing points
        self.L = nn.Parameter(torch.Tensor(1,n_idc,n_rank)) #LL.T = covariance
        nn.init.xavier_uniform_(self.L)
        nn.init.xavier_uniform_(self.y_idc)        
        nn.init.xavier_uniform_(self.x_idc)        
        self.log_alpha = nn.Parameter(torch.ones(1)*0)
        self.log_beta = nn.Parameter(torch.ones(1)*0)
        self.log_gamma = nn.Parameter(torch.ones(1)*0)
        self.n_idc = n_idc

    def rbfkernel(self,z0,z1,alpha=1.0,gamma=1.0):
        return alpha*torch.exp(-0.5/gamma*((z0.unsqueeze(1)-z1.unsqueeze(2))**2).sum(-1))

    def forward(self,x_spt,y_spt,x_qry,y_qry,plot=None):
        device = x_spt.device
        n_task,n_qry,_ = x_qry.shape
        n_spt = x_spt.shape[1]
        alpha = torch.clamp(self.log_alpha,min=-3,max=3).exp()                
        beta = torch.clamp(self.log_beta,min=-3,max=3).exp()                
        gamma = torch.clamp(self.log_gamma,min=-3,max=3).exp()

        x_sptidc = torch.cat((x_spt,self.x_idc.repeat(n_task,1,1)),1)
        y_sptidc = torch.cat((y_spt,self.y_idc.repeat(n_task,1,1)),1)
        n_sptidc = n_spt+self.n_idc
        K_idc = self.rbfkernel(x_sptidc,x_sptidc,alpha,gamma)
        K_idc_I = K_idc+beta*torch.eye(n_sptidc).unsqueeze(0).repeat(n_task,1,1).to(x_spt.device)
        K_qry_idc = self.rbfkernel(x_qry,x_sptidc,alpha,gamma)
        K_qry = self.rbfkernel(x_qry,x_qry,alpha,gamma)        
        K_qry_I = K_qry+beta*torch.eye(n_qry).unsqueeze(0).repeat(n_task,1,1).to(x_spt.device)
        K_idc_inv = torch.inverse(K_idc_I)
        yhat_qry = torch.bmm(K_qry_idc.transpose(1,2),torch.bmm(K_idc_inv,y_sptidc))

        var_qry = alpha+beta-(torch.bmm(K_qry_idc.transpose(1,2),K_idc_inv)*K_qry_idc.transpose(1,2)).sum(-1,keepdim=True)

        n_p = 10
        eps = 1e-5
        ps = torch.arange(1,n_p).to(device)/n_p
        norm = Normal(yhat_qry,var_qry.sqrt())
        cdf_qry = norm.cdf(y_qry)
        sorted_cdf_qry,indices = cdf_qry.sort(1)
        phat = (cdf_qry<=ps.unsqueeze(0).unsqueeze(0)).float().mean(1)
        ps_np = ps.unsqueeze(0).repeat(n_task,1)
        ece = (phat-ps_np).abs().mean()
        err = ((y_qry-yhat_qry)**2).mean()
        log_loss = (0.5*(var_qry+eps).log()+0.5*(y_qry-yhat_qry)**2/var_qry).mean()
        loss = err
        sharp = err

        return loss,err,ece,sharp
    
class MAML(nn.Module):
    def __init__(self,n_att,n_out,n_hiddens=[32,32],dropout=0.0,local_epoch=5,n_p=10,local_lr=1e-2,method=2,n_dropout=30):
        super(MAML,self).__init__()
        self.dropout_training = None
        if method == 2:
            n_att = n_att+1
        elif method == 3:
            n_out = n_out*2
        elif method == 4:
            self.droput_training = True
        hs = [n_att]+n_hiddens+[n_out]
        self.f = crt_net(hs,dropout=dropout)
        self.hs = hs
        self.dropout = dropout
        self.local_epoch = local_epoch
        self.n_p = n_p
        self.local_lr = local_lr
        self.method = method #2:SQR(pinball), 3:ConditonalGaussian(NLL), 4:Dropout
        self.n_dropout = n_dropout
        
    def pinball(self,yhat,y,p):
        return ((yhat-y)*((y<=yhat).float()-p)).mean()

    def loss(self,yhat,y,p):
        if self.method == 2:
            return self.pinball(yhat,y.unsqueeze(2),p)
        elif self.method == 3:
            m,log_std = yhat.chunk(2,-1)
            return log_std.mean()+0.5*(((yhat-y)/log_std.exp())**2).mean()
        elif self.method == 4:
            return ((yhat-y)**2).mean()

    def forward(self,x_spt,y_spt,x_qry,y_qry,plot=None):
        device = x_spt.device
        n_task,n_spt,_ = x_spt.shape
        n_qry = x_qry.shape[1]
        ps = torch.arange(1,self.n_p).to(device)/self.n_p
        ps_numpy = np.arange(1,self.n_p)/self.n_p
        w = OrderedDict(self.named_parameters())
        ps_spt = None
        ps_qry = None
        if self.method == 2:
            ps_spt = ps.unsqueeze(0).unsqueeze(0).unsqueeze(-1).repeat(n_task,n_spt,1,1)
            ps_qry = ps.unsqueeze(0).unsqueeze(0).unsqueeze(-1).repeat(n_task,n_qry,1,1)
            x_spt = torch.cat((x_spt.unsqueeze(2).repeat(1,1,self.n_p-1,1),ps_spt),-1)
            x_qry = torch.cat((x_qry.unsqueeze(2).repeat(1,1,self.n_p-1,1),ps_qry),-1)
        for epoch in range(self.local_epoch):
            yhat_spt = self.adaptation(x_spt,w)
            loss = self.loss(yhat_spt,y_spt,ps_spt)
            grads = torch.autograd.grad(loss,w.values(),create_graph=self.training)
            w = OrderedDict((name,param-self.local_lr*grad) for ((name,param),grad) in zip(w.items(),grads))
        yhat_qry = self.adaptation(x_qry,w)
        loss = self.loss(yhat_qry,y_qry,ps_qry)
        if self.method == 2:
            yrhat_qry = yhat_qry
            yhat_qry = yrhat_qry[:,:,int(self.n_p/2)-1,:]
        elif self.method == 3:
            m,log_std = yhat_qry.chunk(2,-1)
            ppf = torch.FloatTensor(scipy.stats.norm.ppf(ps_numpy)).to(device)
            rhat_qry = log_std.exp().unsqueeze(2)*ppf.unsqueeze(0).unsqueeze(0).unsqueeze(-1)
            yrhat_qry = yhat_qry.unsqueeze(2)+rhat_qry
        elif self.method == 4:
            yhats_qry = self.adaptation(x_qry.unsqueeze(2).repeat(1,1,self.n_dropout,1),w)
            eps = 1e-5
            m = yhats_qry.mean(2)
            std = yhats_qry.std(2)
            ppf = torch.FloatTensor(scipy.stats.norm.ppf(ps_numpy)).to(device)
            rhat_qry = std.unsqueeze(2)*ppf.unsqueeze(0).unsqueeze(0).unsqueeze(-1)
            yrhat_qry = yhat_qry.unsqueeze(2)+rhat_qry            
            
        phat = (yrhat_qry>=y_qry.unsqueeze(2)).float().mean(1)
        ece = (phat-ps.unsqueeze(0).unsqueeze(-1)).abs().mean()
        err = ((y_qry-yhat_qry)**2).mean()
        sharp = torch.relu(yrhat_qry[:,:,-1,:]-yrhat_qry[:,:,0,:]).mean()

        return loss,err,ece,sharp

    def adaptation(self,x,w):
        for i in range(len(self.hs)-1):
            x = F.linear(x,w['f.%d.weight'%(i*3)],w['f.%d.bias'%(i*3)])
            if i != len(self.hs)-2:
                x = F.relu(x)
                if self.method == 4:
                    x = F.dropout(x,self.dropout,training=True)
                else:
                    x = F.dropout(x,self.dropout,training=self.training)                    
        return x

class NeuralProcess(nn.Module):
    def __init__(self,n_att,n_out,n_hiddens=[32,32],dropout=0.0,n_p=10):
        super(NeuralProcess,self).__init__()
        self.n_p = n_p        
        hs = [n_att+n_out]+n_hiddens
        self.g1 = crt_net(hs,dropout=dropout)
        hs = [n_hiddens[-1]]+n_hiddens
        self.g2 = crt_net(hs,dropout=dropout)
        hs = [n_att+n_hiddens[-1]]+n_hiddens+[n_out*2]        
        self.f = crt_net(hs,dropout=dropout)        
        self.dropout = dropout

    def forward(self,x_spt,y_spt,x_qry,y_qry,plot=None):
        n_task,n_spt,_ = x_spt.shape
        n_qry = x_qry.shape[1]
        xy_spt = torch.cat((x_spt,y_spt),2)
        r = self.g2(self.g1(xy_spt).mean(1)) #n_task*n_hidden

        xr_qry = torch.cat((x_qry,r.unsqueeze(1).repeat(1,n_qry,1)),2)
        yhat_qry,ylogstd_qry = torch.chunk(self.f(xr_qry),2,dim=2)
        loss = ylogstd_qry.mean()+0.5*(((yhat_qry-y_qry)/ylogstd_qry.exp())**2).mean()

        ps = torch.arange(1,self.n_p).to(x_spt.device)/self.n_p        
        ps_numpy = np.arange(1,self.n_p)/self.n_p
        ppf = torch.FloatTensor(scipy.stats.norm.ppf(ps_numpy)).to(x_spt.device)
        rhat_qry = ylogstd_qry.exp().unsqueeze(2)*ppf.unsqueeze(0).unsqueeze(0).unsqueeze(-1)
        yrhat_qry = yhat_qry.unsqueeze(2)+rhat_qry
        phat = (yrhat_qry>=y_qry.unsqueeze(2)).float().mean(1)
        ece = (phat-ps.unsqueeze(0).unsqueeze(-1)).abs().mean()
        err = ((y_qry-yhat_qry)**2).mean()
        sharp = torch.relu(yrhat_qry[:,:,-1,:]-yrhat_qry[:,:,0,:]).mean()
        return loss,err,ece,sharp
    
class MetaQuantile(nn.Module):
    def __init__(self,n_att,n_hiddens=[32,32],dropout=0.0,method=0,use_encoder=1,use_gp=0,lamb=0.5,use_loss=0,
                 use_meanfunc=0,n_p=10,use_monotonic=0,use_maml=0,use_taskembed=0,use_weight=1,split_data=0,use_zscore=0,fixed_sigma=0):
        super(MetaQuantile,self).__init__()
        if use_encoder > 0:
            if use_taskembed == 0:
                hs = [n_att]+n_hiddens            
            else:
                hs = [n_att+n_hiddens[-1]]+n_hiddens            
            self.f = crt_net(hs,dropout=dropout)

        self.log_beta_y = nn.Parameter(torch.ones(1)*0)
        self.log_alpha_y = nn.Parameter(torch.ones(1)*0)
        self.log_gamma_y = nn.Parameter(torch.ones(1)*0)
        self.log_sigma = nn.Parameter(torch.ones(1)*0)
        self.log_sigma_qry = torch.ones(1)*-3
        self.log_beta_z = nn.Parameter(torch.ones(1)*0)
        self.log_alpha_z = nn.Parameter(torch.ones(1)*0)
        self.log_gamma_z = nn.Parameter(torch.ones(1)*0)
        self.log_weight_z = nn.Parameter(torch.ones(1)*0)        

        self.fixed_sigma = fixed_sigma
        self.use_zscore = use_zscore
        self.split_data = split_data
        self.use_weight = use_weight
        self.use_encoder = use_encoder
        self.use_gp = use_gp
        self.lamb = lamb
        self.method = method #0:no-recalibration, 1:recalibration, -1:training:no-recalib,test:recalib, -2:recalibration-by-GP -3:recalibration-by-linear
        self.use_loss = use_loss
        self.use_meanfunc = use_meanfunc
        self.n_p = n_p
        self.use_monotonic = use_monotonic
        self.use_maml = use_maml
        self.use_taskembed = use_taskembed
        if use_monotonic == 1:
            hs = [1]+n_hiddens+[1]
            self.monotonic = crt_posnet(hs,dropout=dropout)
        if use_taskembed == 1:
            hs = [n_att]+n_hiddens
            self.taskembed0 = crt_net(hs,dropout=dropout)
            hs = [n_hiddens[-1]]+n_hiddens
            self.taskembed1 = crt_net(hs,dropout=dropout)
        if use_meanfunc == 1:
            if use_taskembed == 0:
                hs = [n_att]+n_hiddens+[1]
            else:
                hs = [n_att+n_hiddens[-1]]+n_hiddens+[1]                
            self.meanfunc = crt_net(hs,dropout=dropout)

    def rbfkernel(self,z0,z1,alpha=1.0,gamma=1.0):
        return alpha*torch.exp(-0.5/gamma*((z0.unsqueeze(1)-z1.unsqueeze(2))**2).sum(-1))
    
    def linrbfkernel(self,z0,z1,alpha=1.0,gamma=1.0,eta=1.0):
        return alpha*torch.exp(-0.5/gamma*((z0.unsqueeze(1)-z1.unsqueeze(2))**2).sum(-1))+eta*(z0.unsqueeze(1)*z1.unsqueeze(2)).sum(-1)

    def pinball(self,yhat,y,p):
        return ((yhat-y)*((y<=yhat).float()-p)).mean()

    def forward(self,x_spt,y_spt,x_qry,y_qry,plot=None):
        if self.split_data == 1:
            n_spt = x_spt.shape[1]
            n_spt_half = int(n_spt/2)
            x_spt_cal = x_spt[:,n_spt_half:,:]
            x_spt = x_spt[:,:n_spt_half,:]
            n_spt_cal = x_spt_cal.shape[1]
            y_spt_cal = y_spt[:,n_spt_half:,:]
            y_spt = y_spt[:,:n_spt_half,:]
            
        n_task,n_spt,n_att = x_spt.shape
        n_task_qry = x_qry.shape[0]
        n_qry = x_qry.shape[1]
        n_all = n_spt+n_qry
        device = x_spt.device

        if self.use_taskembed == 1:
            z_task = self.taskembed1(self.taskembed0(x_spt).mean(1,keepdim=True))
            
        if self.use_encoder == 0:
            z_spt = x_spt
            z_qry = x_qry
            if self.split_data == 1:
                z_spt_cal = x_spt_cal
        else:
            if self.use_taskembed == 1:
                z_spt = self.f(torch.cat((x_spt,z_task.repeat(1,n_spt,1)),-1))
                z_qry = self.f(torch.cat((x_qry,z_task.repeat(1,n_qry,1)),-1))
                if self.split_data == 1:
                    z_spt_cal = self.f(torch.cat((x_spt_cal,z_task.repeat(1,n_spt_cal,1)),-1))                    
            else:
                z_spt = self.f(x_spt)
                z_qry = self.f(x_qry)
                if self.split_data == 1:
                    z_spt_cal = self.f(x_spt_cal)

        if self.use_meanfunc == 1:
            if self.use_taskembed == 1:
                mu_spt = self.meanfunc(torch.cat((x_spt,z_task.repeat(1,n_spt,1)),-1))
                mu_qry = self.meanfunc(torch.cat((x_qry,z_task.repeat(1,n_qry,1)),-1))
                if self.split_data == 1:
                    mu_spt_cal = self.meanfunc(torch.cat((x_spt_cal,z_task.repeat(1,n_spt_cal,1)),-1))
            else:
                mu_spt = self.meanfunc(x_spt)
                mu_qry = self.meanfunc(x_qry)
                if self.split_data == 1:
                    mu_spt_cal = self.meanfunc(x_spt_cal)                

        #fit GP
        alpha = torch.clamp(self.log_alpha_y,min=-3,max=3).exp()        
        gamma = torch.clamp(self.log_gamma_y,min=-3,max=3).exp()        
        beta = torch.clamp(self.log_beta_y,min=-3,max=3).exp()        
        K_spt = self.rbfkernel(z_spt,z_spt,alpha,gamma)
        K_spt_I = K_spt+beta*torch.eye(n_spt).unsqueeze(0).repeat(n_task,1,1).to(x_spt.device)
        K_qry_spt = self.rbfkernel(z_qry,z_spt,alpha,gamma)
        K_spt_inv = torch.inverse(K_spt_I)
        if self.split_data == 1:
            K_cal_spt = self.rbfkernel(z_spt_cal,z_spt,alpha,gamma)
        if self.use_meanfunc == 1:
            yhat_qry = mu_qry+torch.bmm(K_qry_spt.transpose(1,2),torch.bmm(K_spt_inv,y_spt-mu_spt))
            yhat_spt = mu_spt+torch.bmm(K_spt.transpose(1,2),torch.bmm(K_spt_inv,y_spt-mu_spt))
            if self.split_data == 1:
                yhat_spt_cal = mu_spt_cal+torch.bmm(K_cal_spt.transpose(1,2),torch.bmm(K_spt_inv,y_spt-mu_spt))
        else:
            yhat_qry = torch.bmm(K_qry_spt.transpose(1,2),torch.bmm(K_spt_inv,y_spt))
            yhat_spt = torch.bmm(K_spt.transpose(1,2),torch.bmm(K_spt_inv,y_spt))
            if self.split_data == 1:
                yhat_spt_cal = torch.bmm(K_cal_spt.transpose(1,2),torch.bmm(K_spt_inv,y_spt))

        n_p = self.n_p
        ps = torch.arange(1,n_p).to(device)/n_p
        eps = 1e-5
        var_qry = alpha+beta-(torch.bmm(K_qry_spt.transpose(1,2),K_spt_inv)*K_qry_spt.transpose(1,2)).sum(-1,keepdim=True)
        var_spt = alpha+beta-(torch.bmm(K_spt,K_spt_inv)*K_spt).sum(-1,keepdim=True)
        if self.split_data == 1:
            var_spt_cal = alpha+beta-(torch.bmm(K_cal_spt.transpose(1,2),K_spt_inv)*K_cal_spt.transpose(1,2)).sum(-1,keepdim=True)
            
        if self.method == 0 or self.method == -1 and self.training: #no-calibration
            norm_spt = Normal(yhat_spt,var_spt.sqrt())
            cdf_spt = norm_spt.cdf(y_spt)
            norm = Normal(yhat_qry,var_qry.sqrt())
            cdf_qry = norm.cdf(y_qry)
            y_qry975 = norm.icdf(torch.ones((1))*0.975)
            y_qry025 = norm.icdf(torch.ones((1))*0.025)
            zscore_spt = cdf_spt
            zscore_qry = cdf_qry
            uncalib_cdf_qry = cdf_qry
            uncalib_cdf_spt = cdf_spt            
            
        elif self.method == 1 or self.method == -2 or self.method == -3 or self.method == -1 and not self.training: #calibration

            if self.split_data == 1:
                yhat_spt = yhat_spt_cal
                y_spt = y_spt_cal
                var_spt = var_spt_cal
                n_spt = n_spt_cal

            norm_qry = Normal(yhat_qry,var_qry)
            cdf_qry0 = norm_qry.cdf(y_qry)
            norm_spt = Normal(yhat_spt,var_spt)
            cdf_spt0 = norm_spt.cdf(y_spt)
            uncalib_cdf_qry = cdf_qry0
            uncalib_cdf_spt = cdf_spt0
            uncalib_y_qry975 = norm_qry.icdf(torch.ones((1))*0.975)
            uncalib_y_qry025 = norm_qry.icdf(torch.ones((1))*0.025)
            
            eps = 1e-5
            if self.use_zscore == 1:
                zscore_spt = (y_spt-yhat_spt)/(var_spt+eps).sqrt()
                zscore_qry = (y_qry-yhat_qry)/(var_qry+eps).sqrt()
            else:
                zscore_spt = cdf_spt0
                zscore_qry = cdf_qry0

            #recalibration
            #regression by GP sorted-zscore = ranking
            if self.method == -2:
                sorted_zscore_spt,indices = zscore_spt.sort(1)
                zscore_y_spt = ((torch.arange(n_spt))/n_spt).unsqueeze(0).unsqueeze(-1).repeat(n_task,1,1)
                Kz_spt = self.rbfkernel(sorted_zscore_spt,sorted_zscore_spt,self.log_alpha_z.exp(),self.log_gamma_z.exp())
                Kz_spt_I = Kz_spt+self.log_beta_z.exp()*torch.eye(n_spt).unsqueeze(0).repeat(n_task,1,1).to(x_spt.device)
                Kz_qry_spt = self.rbfkernel(zscore_qry,sorted_zscore_spt,self.log_alpha_z.exp(),self.log_gamma_z.exp())
                Kz_spt_spt = self.rbfkernel(zscore_spt,sorted_zscore_spt,self.log_alpha_z.exp(),self.log_gamma_z.exp())
                Kz_spt_inv = torch.inverse(Kz_spt_I)
                cdf_qry = torch.bmm(Kz_qry_spt.transpose(1,2),torch.bmm(Kz_spt_inv,zscore_y_spt))
                cdf_qry = torch.clamp(cdf_qry,0,1)
                cdf_spt = torch.bmm(Kz_spt_spt.transpose(1,2),torch.bmm(Kz_spt_inv,zscore_y_spt))
                cdf_spt = torch.clamp(cdf_spt,0,1)
            elif self.method == -3: #linear
                sorted_zscore_spt,indices = zscore_spt.sort(1)
                zscore_y_spt = (torch.arange(n_spt)/(n_spt-1)).unsqueeze(0).unsqueeze(-1).repeat(n_task,1,1)
                gamma = torch.clamp(self.log_gamma_z,min=-3,max=3).exp()

                Kz_spt = self.rbfkernel(zscore_spt,sorted_zscore_spt,1.0,gamma)
                Kz_qry_spt = self.rbfkernel(zscore_qry,sorted_zscore_spt,1.0,self.log_gamma_z.exp())
                cdf_qry = ((Kz_qry_spt*zscore_y_spt).sum(1)/Kz_qry_spt.sum(1)).unsqueeze(-1)
                cdf_spt = ((Kz_spt*zscore_y_spt).sum(1)/Kz_spt.sum(1)).unsqueeze(-1)
            else: #method==1
                cdf_qry = torch.zeros_like(zscore_qry)
                cdf_spt = torch.zeros_like(zscore_spt)
                gmms = []
                if self.method == -1 or self.fixed_sigma == 1:
                    sigma = torch.ones((1))*1e-8
                else:
                    sigma = torch.clamp(self.log_sigma,min=-3,max=3).exp()
                for i in range(n_task):
                    mix = Categorical(torch.ones(n_spt).to(device))
                    comp = Normal(zscore_spt[i,:,:].squeeze(-1),sigma.repeat(n_spt))
                    gmm = MixtureSameFamily(mix,comp)
                    eps = 0.0
                    if self.method == -1:
                        eps = 1e-5
                    cdf_qry[i,:,:] = gmm.cdf(zscore_qry[i,:,:]+eps)
                    cdf_spt[i,:,:] = gmm.cdf(zscore_spt[i,:,:]+eps)
            if self.use_weight == 1:
                weight_z = torch.sigmoid(self.log_weight_z)
                cdf_qry = weight_z*cdf_qry0+(1-weight_z)*cdf_qry
                cdf_spt = weight_z*cdf_spt0+(1-weight_z)*cdf_spt

            cdf_qry975 = (cdf_qry-0.975).abs().argmin(1).unsqueeze(-1)
            cdf_qry025 = (cdf_qry-0.025).abs().argmin(1).unsqueeze(-1)
            zscore_qry975 = torch.gather(zscore_qry,1,cdf_qry975)
            zscore_qry025 = torch.gather(zscore_qry,1,cdf_qry025)
            if self.use_zscore == 1:
                y_qry975 = zscore_qry975*var_qry.sqrt()+yhat_qry
                y_qry025 = zscore_qry025*var_qry.sqrt()+yhat_qry
            else:
                y_qry975 = norm_qry.icdf(zscore_qry975.clamp(eps,1-eps))
                y_qry025 = norm_qry.icdf(zscore_qry025.clamp(eps,1-eps))

        sharp = (y_qry975-y_qry025).mean()

        sorted_cdf_qry,indices = cdf_qry.sort(1)

        smooth_ece = (sorted_cdf_qry-((torch.arange(n_qry))/n_qry).unsqueeze(0).unsqueeze(-1).repeat(n_task,1,1)).abs().mean()
        
        phat = (cdf_qry<=ps.unsqueeze(0).unsqueeze(0)).float().mean(1)
        ps_np = ps.unsqueeze(0).repeat(n_task,1)
        ece = (phat-ps_np).abs().mean()
        err = ((y_qry-yhat_qry)**2).mean()
        
        if self.use_loss == 0: #qry ece loss
            loss = self.lamb*err+(1-self.lamb)*smooth_ece
        else: #qry logistic loss
            logloss = (0.5*(var_qry+eps).log()+0.5*(y_qry-yhat_qry)**2/var_qry).mean()
            loss = logloss

        if plot is not None:
            import matplotlib as mpl
            mpl.use('Agg')
            import matplotlib.pyplot as plt

            plt.rcParams['pdf.fonttype'] = 42
            plt.rcParams["font.size"] = 12            
            
            print("plot: %s"%(plot))
            os.makedirs(plot,exist_ok=True)

            #recalibration
            #regression by GP sorted-zscore = ranking

            naive_cdf_qry = torch.zeros_like(zscore_qry)
            naive_cdf_spt = torch.zeros_like(zscore_spt)
            sigma = torch.ones((1))*1e-8
            for i in range(n_task):
                mix = Categorical(torch.ones(n_spt).to(device))
                comp = Normal(zscore_spt[i,:,:].squeeze(-1),sigma.repeat(n_spt))
                gmm = MixtureSameFamily(mix,comp)
                eps = 1e-5
                naive_cdf_qry[i,:,:] = gmm.cdf(zscore_qry[i,:,:]+eps)
                naive_cdf_spt[i,:,:] = gmm.cdf(zscore_spt[i,:,:]+eps)
                    
            for i in range(n_task):
                x_spt0 = x_spt[i,:,0]
                y_spt0 = y_spt[i,:,0]
                x_qry0 = x_qry[i,:,0]
                y_qry0 = y_qry[i,:,0]
                std_qry0 = var_qry[i,:,0].sqrt()
                std_spt0 = var_spt[i,:,0].sqrt()
                yhat_qry0 = yhat_qry[i,:,0]

                yhat_qry0_min = y_qry025[i,:,0]
                yhat_qry0_max = y_qry975[i,:,0]
                yhat_spt0 = yhat_spt[i,:,0]                
                zscore_qry0 = zscore_qry[i,:,0]
                zscore_spt0 = zscore_spt[i,:,0]
                cdf_qry0 = cdf_qry[i,:,0]
                cdf_spt0 = cdf_spt[i,:,0]
                naive_cdf_qry0 = naive_cdf_qry[i,:,0]
                naive_cdf_spt0 = naive_cdf_spt[i,:,0]
                uncalib_cdf_qry0 = uncalib_cdf_qry[i,:,0]
                uncalib_cdf_spt0 = uncalib_cdf_spt[i,:,0]
                uncalib_y_qry9750 = uncalib_y_qry975[i,:,0]
                uncalib_y_qry0250 = uncalib_y_qry025[i,:,0]

                if device != 'cpu':
                    x_spt0 = x_spt0.cpu()
                    y_spt0 = y_spt0.cpu()
                    x_qry0 = x_qry0.cpu()
                    y_qry0 = y_qry0.cpu()
                    std_qry0 = std_qry0.cpu()                    
                    std_spt0 = std_spt0.cpu()                    
                    yhat_qry0 = yhat_qry0.cpu()
                    yhat_qry0_min = yhat_qry0_min.cpu()
                    yhat_qry0_max = yhat_qry0_max.cpu()
                    yhat_spt0 = yhat_spt0.cpu()
                    zscore_qry0 = zscore_qry0.cpu()
                    zscore_spt0 = zscore_spt0.cpu()
                    cdf_qry0 = cdf_qry0.cpu()
                    cdf_spt0 = cdf_spt0.cpu()
                    naive_cdf_qry0 = naive_cdf_qry0.cpu()
                    naive_cdf_spt0 = naive_cdf_spt0.cpu()
                    uncalib_cdf_qry0 = uncalib_cdf_qry0.cpu()
                    uncalib_cdf_spt0 = uncalib_cdf_spt0.cpu()
                    uncalib_y_qry9750 = uncalib_y_qry9750.cpu()
                    uncalib_y_qry0250 = uncalib_y_qry0250.cpu()
                    
                x_spt0 = x_spt0.numpy()
                y_spt0 = y_spt0.numpy()
                x_qry0 = x_qry0.numpy()
                y_qry0 = y_qry0.numpy()
                std_qry0 = std_qry0.detach().numpy()                
                std_spt0 = std_spt0.detach().numpy()                
                yhat_qry0 = yhat_qry0.detach().numpy()
                yhat_qry0_min = yhat_qry0_min.detach().numpy()
                yhat_qry0_max = yhat_qry0_max.detach().numpy()
                yhat_spt0 = yhat_spt0.detach().numpy()
                zscore_qry0 = zscore_qry0.detach().numpy()
                zscore_spt0 = zscore_spt0.detach().numpy()
                cdf_qry0 = cdf_qry0.detach().numpy()
                cdf_spt0 = cdf_spt0.detach().numpy()
                naive_cdf_qry0 = naive_cdf_qry0.detach().numpy()
                naive_cdf_spt0 = naive_cdf_spt0.detach().numpy()
                uncalib_cdf_qry0 = uncalib_cdf_qry0.detach().numpy()
                uncalib_cdf_spt0 = uncalib_cdf_spt0.detach().numpy()
                uncalib_y_qry9750 = uncalib_y_qry9750.detach().numpy()
                uncalib_y_qry0250 = uncalib_y_qry0250.detach().numpy()

                plt.clf()
                plt.plot(x_spt0,y_spt0,'ro')
                plt.plot(x_qry0,y_qry0,'r-')
                plt.plot(x_qry0,yhat_qry0,'b-')
                plt.fill_between(x_qry0,yhat_qry0_min,yhat_qry0_max,alpha=0.5,facecolor='b')
                plt.xlabel("x")
                plt.ylabel("y")
                ofn = "%s/plt_i%s.png"%(plot,i)
                plt.savefig(ofn,bbox_inches="tight")
                xlim = plt.xlim()
                ylim = plt.ylim()

                plt.clf()
                plt.plot(x_spt0,y_spt0,'ro')
                plt.plot(x_qry0,y_qry0,'r-')
                plt.plot(x_qry0,yhat_qry0,'b-')
                plt.fill_between(x_qry0,uncalib_y_qry0250,uncalib_y_qry9750,alpha=0.5,facecolor='b')
                plt.xlabel("x")
                plt.ylabel("y")
                plt.xlim(xlim)
                plt.ylim(ylim)
                ofn = "%s/uncalib_plt_i%s.png"%(plot,i)
                plt.savefig(ofn,bbox_inches="tight")
                

                plt.clf()
                plt.plot(zscore_spt0,cdf_spt0,'ro')                
                plt.plot(zscore_qry0,cdf_qry0,'bx')
                plt.xlabel("h")
                plt.ylabel("r(h)")
                ofn = "%s/fcalib_i%s.png"%(plot,i)
                plt.savefig(ofn,bbox_inches="tight")

                plt.clf()
                plt.plot(zscore_spt0,naive_cdf_spt0,'ro')                
                plt.plot(zscore_qry0,naive_cdf_qry0,'bx')
                plt.xlabel("h")
                plt.ylabel("r(h)")
                ofn = "%s/naive_fcalib_i%s.png"%(plot,i)
                plt.savefig(ofn,bbox_inches="tight")                                

                plt.clf()
                uncalib_cdf_spt0 = np.sort(uncalib_cdf_spt0)
                uncalib_cdf_qry0 = np.sort(uncalib_cdf_qry0)
                naive_cdf_spt0 = np.sort(naive_cdf_spt0)
                naive_cdf_qry0 = np.sort(naive_cdf_qry0)
                plt.plot(np.arange(1,n_qry+1)/(n_qry),np.arange(1,n_qry+1)/(n_qry),'k',label="Perfectly calibrated")
                plt.plot(uncalib_cdf_spt0,(np.arange(n_spt)+1)/(n_spt),'g',label="Uncalibrated support")
                plt.plot(naive_cdf_spt0,(np.arange(n_spt)+1)/(n_spt),'m',label="Calibrated support")
                plt.plot(uncalib_cdf_qry0,(np.arange(n_qry)+1)/(n_qry),'b',label="Uncalibrated query")
                plt.plot(naive_cdf_qry0,(np.arange(n_qry)+1)/(n_qry),'r',label="Calibrated query")
                plt.legend()
                plt.xlabel("Estimated CDF")
                plt.ylabel("Empirical CDF")
                ofn = "%s/naive_calib_i%s.png"%(plot,i)
                plt.savefig(ofn,bbox_inches="tight")                                

                plt.clf()
                uncalib_cdf_spt0 = np.sort(uncalib_cdf_spt0)
                uncalib_cdf_qry0 = np.sort(uncalib_cdf_qry0)
                cdf_spt0 = np.sort(cdf_spt0)
                cdf_qry0 = np.sort(cdf_qry0)
                plt.plot(np.arange(1,n_qry+1)/(n_qry),np.arange(1,n_qry+1)/(n_qry),'k',label="Perfectly calibrated")
                plt.plot(uncalib_cdf_spt0,(np.arange(n_spt)+1)/(n_spt),'g',label="Uncalibrated support")
                plt.plot(cdf_spt0,(np.arange(n_spt)+1)/(n_spt),'m',label="Calibrated support")
                plt.plot(uncalib_cdf_qry0,(np.arange(n_qry)+1)/(n_qry),'b',label="Uncalibrated query")
                plt.plot(cdf_qry0,(np.arange(n_qry)+1)/(n_qry),'r',label="Calibrated query")
                plt.legend()
                plt.xlabel("Estimated CDF")
                plt.ylabel("Empirical CDF")
                ofn = "%s/calib_i%s.png"%(plot,i)
                plt.savefig(ofn,bbox_inches="tight")

        return loss,err,ece,sharp
            
