import random
import argparse
import sys, string, glob, re, os
import numpy as np
import time,copy
import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import TensorDataset,DataLoader,Dataset
from torch.nn import functional as F

from model_meta_quantile import MetaQuantile,MAML,SparseGP,NeuralProcess

class IdxDataset(Dataset):
    def __init__(self,idxs):
        self.idxs = idxs
    def __len__(self):
        return len(self.idxs)
    def __getitem__(self, idx):
        return self.idxs[idx]
    
def traineval_epoch(dataloader,xs,ys,model,optimizer,args,is_train,plot=None):
    if is_train == 1:
        model.train()
    else:
        model.eval()

    total_loss = 0.0
    total_err = 0.0
    total_ece = 0.0
    total_sharp = 0.0
    total_cnt = 0
    n_att = args.n_att
    n_spt = args.n_spt
    n_qry = args.n_qry
    n_out = 1
    for batch_idx,(tasks) in enumerate(dataloader):
        n_task = len(tasks)
        x_spt = torch.zeros((n_task,n_spt,n_att)).to(args.device)
        y_spt = torch.zeros((n_task,n_spt,1)).to(args.device)
        x_qry = torch.zeros((n_task,n_qry,n_att)).to(args.device)
        y_qry = torch.zeros((n_task,n_qry,1)).to(args.device)
        for i in range(n_task):
            n_data = xs[tasks[i]].shape[0]
            if is_train == 1:
                perm = np.random.permutation(n_data)
            else:
                perm = np.arange(n_data)                
            x_spt[i,:,:] = torch.FloatTensor(xs[tasks[i]][perm[:n_spt],:]).to(args.device)
            y_spt[i,:,:] = torch.FloatTensor(ys[tasks[i]][perm[:n_spt],:]).to(args.device)
            x_qry[i,:,:] = torch.FloatTensor(xs[tasks[i]][perm[n_spt:n_spt+n_qry],:]).to(args.device)
            y_qry[i,:,:] = torch.FloatTensor(ys[tasks[i]][perm[n_spt:n_spt+n_qry],:]).to(args.device)
        plot0 = None
        if plot is not None:
            plot0 = "%s/b%s/"%(plot,batch_idx)
        loss,err,ece,sharp = model(x_spt,y_spt,x_qry,y_qry,plot=plot0)
        if is_train == 1:
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
        total_cnt += n_task
        total_loss += loss.detach().item()*n_task
        total_err += err.detach().item()*n_task
        total_ece += ece.detach().item()*n_task
        total_sharp += sharp.detach().item()*n_task
            
    return total_loss/total_cnt,total_err/total_cnt,total_ece/total_cnt,total_sharp/total_cnt

def plot_epoch(dataloader,xs,ys,model,optimizer,args,is_train,plot=None):
    model.eval()
    total_loss = 0.0
    total_err = 0.0
    total_ece = 0.0
    total_sharp = 0.0
    total_cnt = 0
    n_att = args.n_att
    n_spt = args.n_spt
    n_out = 1
    for batch_idx,(tasks) in enumerate(dataloader):
        n_task = len(tasks)
        n_data_min = None
        for i in range(n_task):
            n_data = xs[tasks[i]].shape[0]
            if n_data_min is None or n_data_min > n_data:
                n_data_min = n_data
        x_spt = torch.zeros((n_task,n_spt,n_att)).to(args.device)
        y_spt = torch.zeros((n_task,n_spt,1)).to(args.device)
        x_qry = torch.zeros((n_task,n_data,n_att)).to(args.device)
        y_qry = torch.zeros((n_task,n_data,1)).to(args.device)
        for i in range(n_task):
            n_data = xs[tasks[i]].shape[0]
            if is_train == 1:
                perm = np.random.permutation(n_data)
            else:
                perm = np.arange(n_data)                
            x_spt[i,:,:] = torch.FloatTensor(xs[tasks[i]][perm[:n_spt],:]).to(args.device)
            y_spt[i,:,:] = torch.FloatTensor(ys[tasks[i]][perm[:n_spt],:]).to(args.device)
            idxs = np.argsort(xs[tasks[i]][:,0])
            x_qry[i,:,:] = torch.FloatTensor(xs[tasks[i]][idxs,:]).to(args.device)
            y_qry[i,:,:] = torch.FloatTensor(ys[tasks[i]][idxs,:]).to(args.device)
        plot0 = None
        if plot is not None:
            plot0 = "%s/b%s/"%(plot,batch_idx)
        loss,err,ece,sharp = model(x_spt,y_spt,x_qry,y_qry,plot=plot0)
        total_cnt += n_task
        total_loss += loss.detach().item()*n_task
        total_err += err.detach().item()*n_task
        total_ece += ece.detach().item()*n_task
        total_sharp += sharp.detach().item()*n_task
            
    return total_loss/total_cnt,total_err/total_cnt,total_ece/total_cnt,total_sharp/total_cnt


def train(trainloader,validloader,testloader,dat,model,optimizer,args):
    valid_errs = []
    valid_eces = []
    valid_losss = []
    bestmodel = copy.deepcopy(model)

    for epoch_i in range(args.epoch):
        train_loss,train_err,train_ece,train_sharp = traineval_epoch(trainloader,dat['train_xs'],dat['train_ys'],model,optimizer,args,is_train=1)
        valid_loss,valid_err,valid_ece,valid_sharp = traineval_epoch(validloader,dat['valid_xs'],dat['valid_ys'],model,optimizer,args,is_train=0)
        test_loss,test_err,test_ece,test_sharp = traineval_epoch(testloader,dat['test_xs'],dat['test_ys'],model,optimizer,args,is_train=0)
        print('%d train:loss: %f err: %f ece: %f valid:loss: %f err: %f ece: %f test:loss: %f err: %f ece: %f'%(epoch_i,train_loss,train_err,train_ece,valid_loss,valid_err,valid_ece,test_loss,test_err,test_ece))
                                                     
        valid_errs += [valid_err]
        valid_eces += [valid_ece]
        valid_losss += [valid_loss]
        
        if valid_loss <= min(valid_losss):
            bestmodel = copy.deepcopy(model)
            
    if args.out:
        model_state_dict = bestmodel.state_dict()
        checkpoint = {'model': model_state_dict,'settings': args}
        torch.save(checkpoint,"%s.chkpt"%args.out)

    return bestmodel

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--fn',type=str,default="",help='data file, e.g., data/csv/PenDigits_withoutdupl_norm_v01_t100/I10/fns0.npz')
    parser.add_argument('--epoch','-e',type=int,default=1000,help='epochs')
    parser.add_argument('--n_hiddens',type=str,default='32,32',help='#hidden units')
    parser.add_argument('--n_idc',type=int,default=128,help='#inducing points for sparse GP')
    parser.add_argument('--n_spt',type=int,default=10,help='#support sample for each class')
    parser.add_argument('--n_qry',type=int,default=30,help='#query sample for each class')
    parser.add_argument('--r_train',type=float,default=0.7,help='ratio of training tasks')
    parser.add_argument('--r_valid',type=float,default=0.1,help='ratio of validation tasks')

    parser.add_argument('--weight_decay',type=float,default=1e-5)
    parser.add_argument('--dropout', type=float, default=0.0)
    parser.add_argument('--n_dropout', type=int, default=30)    
    parser.add_argument('--method',type=int,default=1,help='0:no-calibration 1:calibration')
    parser.add_argument('--seed',type=int,default=0,help='random seed')
    parser.add_argument('--local_epoch',type=int,default=5,help='MAML inner epoch')
    parser.add_argument('--local_lr',type=float,default=1e-2,help='MAML inner lr')
    parser.add_argument('--batch_size', type=int, default=32)
    parser.add_argument('--no_cuda', action='store_true')
    parser.add_argument('--lr', type=float, default=1e-2)
    parser.add_argument('--initmodel', default=None)
    parser.add_argument('--out', default=None)
    parser.add_argument('--use_encoder',type=int,default=1,help='use support set for neural net (1) or not (0)')
    parser.add_argument('--use_weight',type=int,default=1,help='use weight for recalibrated and uncalibrated cdf')
    parser.add_argument('--use_zscore',type=int,default=0,help='use zscore for calibration')
    parser.add_argument('--fixed_sigma',type=int,default=0,help='fix sigma for kernel density estimation')
    parser.add_argument('--split_data',type=int,default=0,help='split support for adaptation and recalibration')
    parser.add_argument('--n_dataset',type=int,default=0,help='number of tasks for training')
    parser.add_argument('--lamb',type=float,default=0.5,help="prediction error weight [0,1]")
    parser.add_argument('--use_gp',type=int,default=1,help='0:linear 1:GP')
    parser.add_argument('--use_meanfunc',type=int,default=1,help='0:no 1:use mean function for GP')    
    parser.add_argument('--use_monotonic',type=int,default=0,help='0:no 1:use monotonic function')    
    parser.add_argument('--use_loss',type=int,default=0,help='0:MSE+ece 1:piball 2:logloss')
    parser.add_argument('--use_maml',type=int,default=0,help='0:no 1:maml')
    parser.add_argument('--use_taskembed',type=int,default=0,help='0:no 1:maml')    
    parser.add_argument('--encoder',type=int,default=1,help='0:no-encoder 1:neural net')
    parser.add_argument('--plot',type=str,default=None,help='plot dir')
                        
    args = parser.parse_args()    
    args.cuda = not args.no_cuda and torch.cuda.is_available()
    device = torch.device("cuda" if not args.no_cuda and torch.cuda.is_available() else "cpu")
    args.device = device
    np.random.seed(args.seed)
    random.seed(args.seed)
    torch.manual_seed(args.seed)

    print(args)

    dat = np.load(args.fn,allow_pickle=True)
    n_train = len(dat['train_xs'])
    n_valid = len(dat['valid_xs'])
    n_test = len(dat['test_xs'])
    n_att = dat['train_xs'][0].shape[1]
    n_out = dat['train_ys'][0].shape[1]
    args.n_att = n_att
    if args.n_dataset > 0 and n_train > args.n_dataset:
        n_train = args.n_dataset
        dat_train_xs = dat['train_xs']
        dat_train_xs = dat_train_xs[:n_train]

    args.n_hiddens = list(map(int,args.n_hiddens.split(',')))

    trainidxs = IdxDataset(np.arange(n_train))
    trainloader = DataLoader(trainidxs,batch_size=args.batch_size,shuffle=True)
    valididxs = IdxDataset(np.arange(n_valid))
    validloader = DataLoader(valididxs,batch_size=args.batch_size,shuffle=False)
    testidxs = IdxDataset(np.arange(n_test))
    testloader = DataLoader(testidxs,batch_size=args.batch_size,shuffle=False)

    print("#train: %d #valid: %d #test: %d #att: %d #out: %d"%(n_train,n_valid,n_test,n_att,n_out))

    start = time.time()
    if args.method == 0 or args.method == 1 or args.method == -1 or args.method == -2 or args.method == -3:
        model = MetaQuantile(n_att=n_att,n_hiddens=args.n_hiddens,dropout=args.dropout,
                             method=args.method,lamb=args.lamb,
                             use_gp=args.use_gp,use_loss=args.use_loss,use_encoder=args.use_encoder,use_meanfunc=args.use_meanfunc,
                             use_monotonic=args.use_monotonic,use_maml=args.use_maml,use_taskembed=args.use_taskembed,
                             use_weight=args.use_weight,split_data=args.split_data,use_zscore=args.use_zscore,fixed_sigma=args.fixed_sigma)
    elif args.method == 2 or args.method == 3 or args.method == 4:
        model = MAML(n_att=n_att,n_out=n_out,n_hiddens=args.n_hiddens,dropout=args.dropout,
                     local_epoch=args.local_epoch,local_lr=args.local_lr,method=args.method,n_dropout=args.n_dropout)
    elif args.method == 5:
        model = SparseGP(n_att=n_att,n_out=n_out,n_idc=args.n_idc)
    elif args.method == 6:
        model = NeuralProcess(n_att=n_att,n_out=n_out,n_hiddens=args.n_hiddens,dropout=args.dropout)
        
    model.to(device)
    if args.initmodel:
        model.load_state_dict(torch.load("%s.chkpt"%args.initmodel)['model'],strict=False)

    optimizer = None
    if args.epoch > 0:
        optimizer = torch.optim.Adam(model.parameters(),lr=args.lr,weight_decay=args.weight_decay)    
        model = train(trainloader,validloader,testloader,dat,model,optimizer,args)

    train_time = (time.time()-start)
    start = time.time()
    test_loss,test_err,test_ece,test_sharp = traineval_epoch(testloader,dat['test_xs'],dat['test_ys'],model,optimizer,args,is_train=0)
    if args.plot is not None:
        plot_epoch(testloader,dat['test_xs'],dat['test_ys'],model,optimizer,args,is_train=0,plot=args.plot)    
    test_time = (time.time()-start)

    print("train:time: %f test:time: %f loss: %f err: %f ece: %f sharp: %f"%(train_time,test_time,test_loss,test_err,test_ece,test_sharp))

    
